#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <vector>
#include <algorithm>

using namespace std;

typedef void (*recurseFunctionType)(void *workParameter, char *directoryName, char *fileName);

char MUSIC_HEADER[] = "a:\\Music";
unsigned char BLOCK_HEADER[] = { 0xBE, 0xBA, 0x02, 0x00, 0x01 };
const int FILENAME_OFFSET = 0x144;

void browsePath(void *workParameter, char *startPath, char *mask, DWORD fileAttr, recurseFunctionType recurseFunction);

void addBlock(FILE *blockFile, char *fileName) {
	unsigned char block[716];
	memset(block, 0, sizeof(block));
	memcpy(block, BLOCK_HEADER, sizeof(BLOCK_HEADER));
	sprintf((char*)block + FILENAME_OFFSET, "%s\\%s", MUSIC_HEADER, fileName);
	fwrite(block, 1, sizeof(block), blockFile);
}	

void addDirectoryName(void *workParameter, char *directoryName, char *fileName) {
	printf("Adding directory %s\n", fileName);
	addBlock((FILE*)workParameter, fileName);
}

void addFile(void *workParameter, char *directoryName, char *fileName) {
	char filePath[MAX_PATH];
	sprintf(filePath, "%s\\%s", directoryName, fileName);
	((vector<string>*)workParameter)->push_back(filePath);
}

void checkFile(void *workParameter, char *directoryName, char *fileName) {
	sprintf((char*)workParameter, "%s\\%s", directoryName, fileName);
}

void addM3U(FILE *fileList, char *M3UPath) {
	FILE *M3UFile;
	char M3ULine[MAX_PATH];
	printf("\tParsing M3U %s\n", M3UPath);
	M3UFile = fopen(M3UPath, "r");
	while(!feof(M3UFile)) {
		FILE *mp3File;
		char mp3Path[MAX_PATH];
		fgets(M3ULine, sizeof(M3ULine), M3UFile);
		M3ULine[strlen(M3ULine) - 1] = '\0';
		if (M3ULine[0] == '#' || strlen(M3ULine) < 4)
			continue; // skip comment
		// Check if the file actually exists
		strcpy(mp3Path, M3UPath);
		*(strrchr(mp3Path, '\\') + 1) = '\0';
		strcat(mp3Path, M3ULine);
		mp3File = fopen(mp3Path, "rb");
		if (mp3File) {
			fclose(mp3File);
			printf("\tAdding %s\n", mp3Path);
			addBlock(fileList, mp3Path);			
		}
	}
	fclose(M3UFile);
}

void addDirectoryList(void *workParameter, char *directoryName, char *fileName) {
	FILE *fileList;
	char listPath[MAX_PATH];
	char M3UFile[MAX_PATH];
	printf("Browsing directory %s\n", fileName);
	sprintf(listPath, "%s\\Default Playlist.epl", fileName);
	fileList = fopen(listPath, "wb");
	// Check if we have a m3u file
	M3UFile[0] = '\0';
	browsePath((void*)M3UFile, fileName, "*.m3u", 0, checkFile);
	if (strlen(M3UFile))
		addM3U(fileList, M3UFile);	
	else {
		vector<string> fileNames;
		int i;
		int size;
		printf("\tM3U file not found, adding individual files\n");		
		browsePath((void*)&fileNames, fileName, "*.mp3", 0, addFile);
		sort(fileNames.begin(), fileNames.end());
		size = fileNames.size();
		for (i=0; i<size; i++) {
			printf("\tAdding MP3 file %s\n", fileNames[i].c_str());
			addBlock(fileList, (char*)fileNames[i].c_str());
		}
	}
	fclose(fileList);
}


void browsePath(void *workParameter, char *startPath, char *mask, DWORD fileAttr, recurseFunctionType recurseFunction) {
	char workPath[MAX_PATH];
	WIN32_FIND_DATA result;
	HANDLE findHandle;

	sprintf(workPath, "%s\\%s", startPath, mask);
	findHandle = FindFirstFile(workPath, &result);
	if (findHandle == INVALID_HANDLE_VALUE)
		return;
	if ((!fileAttr || result.dwFileAttributes & fileAttr) && (strcmp(result.cFileName, ".") != 0 && strcmp(result.cFileName, "..") != 0))
		recurseFunction(workParameter, startPath, result.cFileName);
	while (FindNextFile(findHandle, &result)) {
		if ((!fileAttr || result.dwFileAttributes & fileAttr) && (strcmp(result.cFileName, ".") != 0 && strcmp(result.cFileName, "..") != 0))
			recurseFunction(workParameter, startPath, result.cFileName);
	}
}

int main(int argc, char **argv) {
	FILE *fileList;	

	fileList = fopen("Default Playlist.epl", "wb");
	// Write directories list
	browsePath((void*)fileList, ".\\", "*", FILE_ATTRIBUTE_DIRECTORY, addDirectoryName);	
	fclose(fileList);
	// Recurse directories
	browsePath(NULL, ".\\", "*", FILE_ATTRIBUTE_DIRECTORY, addDirectoryList);	
}
